"""
The main entrypoint for running a dataset config
"""

from datetime import datetime

from anystore.types import Uri
from pydantic import BaseModel

from investigraph.model.config import Config, get_config
from investigraph.model.context import DatasetContext


class WorkflowRun(BaseModel):
    """
    Defines a workflow run
    """

    config: Config
    start: datetime
    end: datetime
    entities_uri: Uri | None
    index_uri: Uri | None
    statistics_uri: Uri | None


def run(
    config_uri: Uri,
    store_uri: Uri | None = None,
    entities_uri: Uri | None = None,
    index_uri: Uri | None = None,
    statistics_uri: Uri | None = None,
) -> WorkflowRun:
    start = datetime.now()
    config = get_config(config_uri)
    config.load.uri = store_uri or config.load.uri
    config.export.index_uri = index_uri or config.export.index_uri
    config.export.entities_uri = entities_uri or config.export.entities_uri
    config.export.statistics_uri = statistics_uri or config.export.statistics_uri
    ctx = DatasetContext(config=config)

    has_run = False  # if all sources are cached
    for sctx in ctx.get_sources():
        records = sctx.extract()
        proxies = sctx.transform(records)
        sctx.load(proxies)
        has_run = True

    if has_run:
        ctx.export()
    else:
        ctx.log.info("No data exported as all sources are cached.")

    end = datetime.now()
    return WorkflowRun(
        start=start,
        end=end,
        config=config,
        entities_uri=config.export.entities_uri,
        index_uri=config.export.index_uri,
        statistics_uri=config.export.statistics_uri,
    )



from typing import TYPE_CHECKING

from anystore.types import Uri
from followthemoney import StatementEntity
from ftmq.types import StatementEntities

from investigraph.model.context import get_source_context
from investigraph.model.mapping import QueryMapping

if TYPE_CHECKING:
    from investigraph.model import SourceContext

from ftmq.util import make_entity

from investigraph.types import Record


def map_record(
    record: Record, mapping: QueryMapping, dataset: str | None = "default"
) -> StatementEntities:
    _mapping = mapping.get_mapping()
    if _mapping.source.check_filters(record):
        entities = _mapping.map(record)
        for proxy in entities.values():
            # Use new ftmq API: make_entity(data, entity_type, default_dataset)
            yield make_entity(proxy.to_dict(), StatementEntity, dataset)


def map_ftm(ctx: "SourceContext", record: Record, ix: int) -> StatementEntities:
    """
    The default handler for the transform stage. It takes a
    [Mapping](https://followthemoney.tech/docs/mappings/) and executes it on
    each incoming record.

    Args:
        ctx: instance of the current `SourceContext`
        record: The record to transform, it is an arbitrary `dict[str, Any]`
        ix: The 1-based index of this record (e.g. line number of the extracted
            source)

    Yields:
        Generator of `StatementEntity` instances
    """
    for mapping in ctx.config.transform.queries:
        yield from map_record(record, mapping, ctx.config.dataset.name)


def transform_record(config_uri: Uri, record: Record, ix: int) -> StatementEntities:
    sctx = get_source_context(config_uri, record.get("__source__", "stdin"), uri="-")
    yield from sctx.config.transform.handle(sctx, record, ix)



from followthemoney.proxy import EntityProxy

from investigraph.logic.export import proxy_merge
from investigraph.model import Source, SourceContext, TaskContext
from investigraph.settings import VERSION as __version__

# FIXME overwrite legacy merge with our downgrading merge for ftmstore:
EntityProxy.merge = proxy_merge

__all__ = [
    "__version__",
    "SourceContext",
    "TaskContext",
    "Source",
]



from typing import Annotated, Optional, TypeAlias

import typer
from anystore.cli import ErrorHandler
from anystore.io import (
    IOFormat,
    smart_stream_data,
    smart_stream_json_models,
    smart_write,
    smart_write_data,
    smart_write_models,
)
from anystore.logging import configure_logging, get_logger
from anystore.types import Uri
from followthemoney import StatementEntity
from ftmq.io import smart_read_proxies, smart_write_proxies
from rich import print

from investigraph.exceptions import ImproperlyConfigured
from investigraph.inspect import inspect_config
from investigraph.logic.transform import transform_record
from investigraph.model.context import get_dataset_context, get_source_context
from investigraph.model.source import Source
from investigraph.pipeline import run
from investigraph.settings import VERSION, Settings

settings = Settings()
cli = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=settings.debug)
log = get_logger(__name__)

CONFIG_URI: TypeAlias = Annotated[
    str | None,
    typer.Option("-c", help="Any local or remote json or yaml uri"),
]


class ConfigUri(ErrorHandler):
    def __init__(self, config_uri: Uri | None = None, *args, **kwargs):
        uri = config_uri or settings.config
        if not uri:
            raise ImproperlyConfigured("Specify config.yml either via `-c` flag or ENV")
        self.config_uri = uri
        super().__init__(*args, **kwargs)

    def __enter__(self):
        return self.config_uri


@cli.callback(invoke_without_command=True)
def cli_version(
    version: Annotated[Optional[bool], typer.Option(..., help="Show version")] = False,
):
    if version:
        print(VERSION)
        raise typer.Exit()
    configure_logging()


@cli.command("run")
def cli_run(
    config: CONFIG_URI = None,
    store_uri: Annotated[Optional[str], typer.Option(...)] = None,
    index_uri: Annotated[Optional[str], typer.Option(...)] = None,
    entities_uri: Annotated[Optional[str], typer.Option(...)] = None,
    statistics_uri: Annotated[Optional[str], typer.Option(...)] = None,
):
    """
    Execute a dataset pipeline
    """
    with ConfigUri(config) as config_uri:
        print(
            run(
                config_uri,
                store_uri=store_uri,
                entities_uri=entities_uri,
                index_uri=index_uri,
                statistics_uri=statistics_uri,
            )
        )


@cli.command("seed")
def cli_seed(
    config: CONFIG_URI = None,
    out_uri: Annotated[str, typer.Option("-o")] = "-",
    output_format: Annotated[IOFormat, typer.Option()] = IOFormat.json,
    limit: Annotated[
        int | None,
        typer.Option("-l", help="Only get this number of items"),
    ] = None,
):
    """
    Execute a dataset pipelines seed stage and write sources to out_uri
    (default: stdout)
    """
    with ConfigUri(config) as config_uri:
        ctx = get_dataset_context(config_uri)
        sources = (c.source for c in ctx.get_sources(limit))
        smart_write_models(out_uri, sources, output_format=output_format.name)


@cli.command("extract")
def cli_extract(
    config: CONFIG_URI = None,
    source: Annotated[
        str | None, typer.Option("-s", help="Source name (from config)")
    ] = None,
    from_stdin: Annotated[bool, typer.Option()] = False,
    out_uri: Annotated[str, typer.Option("-o")] = "-",
    output_format: Annotated[IOFormat, typer.Option()] = IOFormat.json,
    limit: Annotated[
        int | None,
        typer.Option("-l", help="Only get this number of items"),
    ] = None,
):
    """
    Execute a dataset pipelines extract stage and write records to out_uri
    (default: stdout)
    """
    with ConfigUri(config) as config_uri:
        if from_stdin:
            for source_ in smart_stream_json_models("-", Source):
                if source is None or source == source_.name:
                    ctx = get_source_context(config_uri, source_.name, uri=source_.uri)
                    smart_write_data(
                        out_uri, ctx.extract(limit), output_format=output_format.name
                    )
        else:
            extractor = []
            ctx = get_dataset_context(config_uri)
            if source is not None:
                for sctx in ctx.get_sources():
                    if sctx.source.name == source:
                        extractor = sctx.extract(limit)
                        break
            else:
                extractor = ctx.extract_all(limit)
            smart_write_data(out_uri, extractor, output_format=output_format.name)


@cli.command("transform")
def cli_transform(
    config: CONFIG_URI = None,
    in_uri: Annotated[str, typer.Option("-i")] = "-",
    out_uri: Annotated[str, typer.Option("-o")] = "-",
    input_format: Annotated[IOFormat, typer.Option()] = IOFormat.json,
):
    """
    Execute a dataset pipelines transform stage with records from in_uri
    (default: stdin) and write proxies to out_uri (default: stdout)
    """
    with ConfigUri(config) as config_uri:

        def _proxies():
            for ix, record in enumerate(
                smart_stream_data(in_uri, input_format=input_format.name)
            ):
                yield from transform_record(config_uri, record, ix)

        smart_write_proxies(out_uri, _proxies())


@cli.command("load")
def cli_load(
    config: CONFIG_URI = None,
    in_uri: Annotated[str, typer.Option("-i")] = "-",
):
    """
    Execute a dataset pipelines load stage with proxies from in_uri
    (default: stdin)
    """
    with ConfigUri(config) as config_uri:
        proxies = smart_read_proxies(in_uri, entity_type=StatementEntity)
        ctx = get_dataset_context(config_uri)
        ctx.load(proxies)


@cli.command("inspect")
def cli_inspect(config: CONFIG_URI = None):
    """Validate dataset config"""
    with ConfigUri(config) as config_uri:
        result = inspect_config(config_uri)
        print(f"[bold green]OK[/bold green] `{config}`")
        print(f"[bold]dataset:[/bold] {result.dataset.name}")
        print(f"[bold]title:[/bold] {result.dataset.title}")


@cli.command("settings")
def cli_settings(
    out_uri: Annotated[str, typer.Option("-o")] = "-",
):
    """
    Show current settings
    """
    with ErrorHandler():
        if out_uri == "-":
            print(settings)
        else:
            smart_write(out_uri, settings.model_dump_json().encode())



from anystore.model import Stats
from anystore.store.resource import UriResource
from anystore.util import ensure_uri
from normality import slugify
from pydantic import BaseModel
from runpandarun import Playbook
from runpandarun.util import PathLike, absolute_path


class Source(BaseModel):
    """
    A model describing an arbitrary local or remote source.
    """

    name: str
    """Identifier of the source (defaults to slugified uri)"""

    uri: str
    """Local or remote uri of this source (via `anystore` / `fsspec`)"""

    pandas: Playbook | None = None
    """Pandas transformation spec (via `runpandarun`)"""

    data: dict | None = {}
    """Arbitrary extra data"""

    def __init__(self, **data):
        data["name"] = data.get("name", slugify(data["uri"]))
        super().__init__(**data)
        self._info: Stats | None = None

    @property
    def resource(self) -> UriResource:
        return UriResource(self.uri)

    @property
    def mimetype(self) -> str:
        return self.resource.info().mimetype

    def info(self) -> Stats:
        if self._info is None:
            self._info = self.resource.info()
        return self._info

    def ensure_uri(self, base: PathLike) -> None:
        """
        ensure absolute file paths based on base path of parent config.yml
        """
        uri = self.uri
        if self.resource.is_local:
            uri = absolute_path(uri, base)
        self.uri = ensure_uri(uri)



from typing import TYPE_CHECKING, Any, Callable, ClassVar, TypeAlias

from anystore.types import Uri
from anystore.util import pydantic_merge
from banal import ensure_list, keys_values
from pydantic import BaseModel
from runpandarun import Playbook

from investigraph.model.mapping import QueryMapping
from investigraph.settings import Settings
from investigraph.util import get_func

if TYPE_CHECKING:
    from investigraph.model.context import SourceContext, DatasetContext

from investigraph.model.source import Source

settings = Settings()

CTX: TypeAlias = "SourceContext | DatasetContext"


class Stage(BaseModel):
    default_handler: ClassVar[str] = ""
    handler: str = ""

    def __init__(self, **data):
        data["handler"] = data.pop("handler", self.default_handler)
        super().__init__(**data)

    def get_handler(self) -> Callable:
        return get_func(self.handler)

    def handle(self, ctx: CTX, *args, **kwargs) -> Any:
        handler = self.get_handler()
        return handler(ctx, *args, **kwargs)


class SeedStage(Stage):
    default_handler = settings.seeder

    uri: str | None = None
    """Base uri for sources"""

    prefix: str | None = None
    """Only include sources with given name prefix"""

    exclude_prefix: str | None = None
    """Exclude sources with given name prefix"""

    glob: str | list[str] | None = None
    """Only include sources that match this glob pattern(s)"""

    storage_options: dict[str, Any] | None = None
    """Pass through kwargs to `fsspec`"""

    source_options: dict[str, Any] | None = None
    """Pass through extra data to source object"""


class ExtractStage(Stage):
    default_handler = settings.extractor

    sources: list[Source] = []
    pandas: Playbook = Playbook()

    def __init__(self, **data):
        super().__init__(**data)
        for source in self.sources:
            if source.pandas is not None:
                source.pandas = pydantic_merge(self.pandas, source.pandas)
            else:
                source.pandas = source.pandas or self.pandas


class TransformStage(Stage):
    default_handler = settings.transformer

    queries: list[QueryMapping] = []

    def __init__(self, **data):
        data["queries"] = ensure_list(keys_values(data, "queries", "query"))
        super().__init__(**data)


class LoadStage(Stage):
    default_handler = settings.loader

    uri: str = settings.store_uri


class ExportStage(Stage):
    default_handler = settings.exporter

    index_uri: Uri | None = None
    entities_uri: Uri | None = None
    statistics_uri: Uri | None = None



"""
aggregate fragments to export
"""

from typing import TYPE_CHECKING

from anystore.io import logged_items, smart_write_model
from anystore.logging import get_logger
from followthemoney import StatementEntity
from followthemoney.proxy import E
from ftmq.aggregate import merge
from ftmq.io import smart_write_proxies
from ftmq.model import Dataset
from ftmq.model.stats import Collector
from ftmq.types import StatementEntities
from ftmq.util import make_entity

if TYPE_CHECKING:
    from investigraph.model import DatasetContext

log = get_logger(__name__)


def proxy_merge(self: E, other: E) -> E:
    """
    Used to override `EntityProxy.merge` in `investigraph.__init__.py`
    """
    return merge(
        make_entity(self.to_dict(), StatementEntity),
        make_entity(other.to_dict(), StatementEntity),
        downgrade=True,
    )


def get_iterator(proxies: StatementEntities, collector: Collector) -> StatementEntities:
    for proxy in logged_items(proxies, "Export", item_name="Proxy", logger=log):
        collector.collect(proxy)
        yield proxy


def handle(ctx: "DatasetContext", *args, **kwargs) -> Dataset:
    """
    The default handler of the export stage. It iterates through the entities
    store, calculates dataset statistics and writes the entities and dataset
    index to json files.

    If neither `entities_uri` or `index_uri` is set, no stats for the `Dataset`
    are computed.

    Args:
        ctx: The current runtime `DatasetContext`

    Returns:
        The `Dataset` object with calculated statistics.
    """
    collector = Collector()
    proxies = ctx.store.iterate(dataset=ctx.dataset)
    iterator = get_iterator(proxies, collector)
    if ctx.config.export.entities_uri:
        smart_write_proxies(ctx.config.export.entities_uri, iterator)
    elif ctx.config.export.index_uri:
        # still compute statistics by iterating through the proxy iterator
        _ = [p for p in iterator]

    if ctx.config.export.index_uri or ctx.config.export.statistics_uri:
        stats = collector.export()
        ctx.config.dataset.apply_stats(stats)
        if ctx.config.export.index_uri:
            smart_write_model(ctx.config.export.index_uri, ctx.config.dataset)
        if ctx.config.export.statistics_uri:
            smart_write_model(ctx.config.export.statistics_uri, stats)

    return ctx.config.dataset



from functools import cache
from os import PathLike
from pathlib import Path
from typing import Self
from urllib.parse import urlparse

from anystore.model.base import BaseModel
from anystore.types import Uri
from anystore.util import ensure_uri
from ftmq.model import Dataset
from pydantic import ConfigDict
from runpandarun.util import absolute_path

from investigraph.model.stage import (
    ExportStage,
    ExtractStage,
    LoadStage,
    SeedStage,
    TransformStage,
)
from investigraph.settings import Settings
from investigraph.util import is_module

settings = Settings()


class Config(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    dataset: Dataset
    base_path: Path = Path()
    seed: SeedStage = SeedStage()
    extract: ExtractStage = ExtractStage()
    transform: TransformStage = TransformStage()
    load: LoadStage = LoadStage()
    export: ExportStage = ExportStage()

    def __init__(self, **data):
        if "dataset" not in data:
            data["dataset"] = data
        super().__init__(**data)
        # ensure absolute file paths for local sources
        self.base_path = Path(self.base_path).absolute()
        for source in self.extract.sources:
            source.ensure_uri(self.base_path)

    @classmethod
    def from_uri(cls, uri: Uri, base_path: PathLike | None = None) -> Self:
        if base_path is None:
            u = urlparse(str(uri))
            if not u.scheme or u.scheme == "file":
                base_path = Path(uri).absolute().parent
        config = cls._from_uri(uri, base_path=base_path)

        # custom user code
        if not is_module(config.seed.handler):
            config.seed.handler = str(
                absolute_path(config.seed.handler, config.base_path)
            )
        if not is_module(config.extract.handler):
            config.extract.handler = str(
                absolute_path(config.extract.handler, config.base_path)
            )
        if not is_module(config.transform.handler):
            config.transform.handler = str(
                absolute_path(config.transform.handler, config.base_path)
            )
        if not is_module(config.load.handler):
            config.load.handler = str(
                absolute_path(config.load.handler, config.base_path)
            )
        if not is_module(config.export.handler):
            config.export.handler = str(
                absolute_path(config.export.handler, config.base_path)
            )

        # ensure base export uris when using memory store
        if config.load.uri.startswith("memory"):
            if config.export.entities_uri is None:
                config.export.entities_uri = ensure_uri(
                    settings.data_root / config.dataset.name / "entities.ftm.json"
                )
            if config.export.index_uri is None:
                config.export.index_uri = ensure_uri(
                    settings.data_root / config.dataset.name / "index.json"
                )

        return config


@cache
def get_config(
    uri: Uri, index_uri: Uri | None = None, entities_uri: Uri | None = None
) -> Config:
    config = Config.from_uri(uri)
    config.export.index_uri = index_uri or config.export.index_uri
    config.export.entities_uri = entities_uri or config.export.entities_uri
    return config



from typing import TYPE_CHECKING, Iterable

from followthemoney import StatementEntity

if TYPE_CHECKING:
    from investigraph.model import DatasetContext


def handle(ctx: "DatasetContext", proxies: Iterable[StatementEntity]) -> int:
    """
    The default handler for the load stage. It writes the given proxies to the
    configured store.

    Args:
        ctx: instance of the current runtime `DatasetContext`
        proxies: Iterable of `StatementEntity`

    Returns:
        The number of entities written to the store.
    """
    ix = 0
    with ctx.store.writer() as bulk:
        for proxy in proxies:
            bulk.add_entity(proxy)
            ix += 1
    return ix



from investigraph.model.config import Config
from investigraph.model.context import DatasetContext, SourceContext, TaskContext
from investigraph.model.source import Source

__all__ = [
    "Config",
    "DatasetContext",
    "Source",
    "SourceContext",
    "TaskContext",
]



from typing import Any, Generator, TypeAlias

Record: TypeAlias = dict[str, Any]
"""A string keyed dict"""

RecordGenerator: TypeAlias = Generator[Record, None, None]
"""A generator of [records][investigraph.types.Record]"""



from functools import cache, cached_property
from typing import IO, Any, AnyStr, ContextManager, Generator

from anystore import smart_open
from anystore.io import logged_items
from anystore.logging import get_logger
from anystore.logic.constants import DEFAULT_MODE
from anystore.store import Store as AnyStore
from anystore.types import Uri
from followthemoney import StatementEntity
from followthemoney.util import make_entity_id
from ftmq.aggregate import merge
from ftmq.model import Dataset
from ftmq.store import Store as FtmStore
from ftmq.store import get_store as get_ftm_store
from ftmq.types import StatementEntities
from ftmq.util import join_slug, make_fingerprint_id
from memorious.logic.fetch import fetch
from pydantic import BaseModel, ConfigDict
from structlog.stdlib import BoundLogger

from investigraph.exceptions import DataError
from investigraph.model.config import Config, get_config
from investigraph.model.source import Source
from investigraph.settings import Settings
from investigraph.types import RecordGenerator
from investigraph.util import make_entity

settings = Settings()


class DatasetContext(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    config: Config

    @property
    def dataset(self) -> str:
        """The dataset name (identifier)"""
        return self.config.dataset.name

    @property
    def prefix(self) -> str:
        """The dataset id prefix (defaults to its name)"""
        return self.config.dataset.prefix or self.dataset

    @cached_property
    def cache(self) -> AnyStore:
        """A shared cache instance"""
        return settings.cache.to_store()

    @cached_property
    def store(self) -> FtmStore:
        """The statement store instance to write fragments to"""
        return get_ftm_store(self.config.load.uri, dataset=self.config.dataset.name)

    @property
    def log(self) -> BoundLogger:
        """A structlog dataset logging instance for the runtime"""
        return get_logger(f"investigraph.datasets.{self.dataset}")

    def extract_all(self, limit: int | None = None) -> RecordGenerator:
        """
        Extract all records from all sources.

        Args:
            limit: Optionally only return this number of items per source (for
                debugging purposes)

        Yields:
            Generator of dictionaries `dict[str, Any]` that are the extracted records.
        """
        for ctx in self.get_sources():
            for ix, rec in enumerate(ctx.extract(), 1):
                if limit is not None and ix > limit:
                    break
                yield rec

    def load(self, proxies: StatementEntities, *args, **kwargs) -> int:
        """
        Load transformed records with the configured handler.
        Defaults to [`investigraph.logic.load:handle`][investigraph.logic.load]

        Args:
            proxies: Generator of StatementEntity instances

        Returns:
            Number of entities loaded to store
        """
        proxies = logged_items(
            proxies,
            "Load",
            item_name="Proxy",
            logger=self.log,
            dataset=self.dataset,
            store=self.config.load.uri,
        )
        return self.config.load.handle(self, proxies, *args, **kwargs)

    def export(self, *args, **kwargs) -> Dataset:
        """
        Execute the configured export handler.
        Defaults to [`investigraph.logic.export:handle`][investigraph.logic.export]

        Returns:
            Dataset model instance for the pipeline dataset, with computed stats
                if configured.
        """
        return self.config.export.handle(self, *args, **kwargs)

    def get_sources(
        self, limit: int | None = None
    ) -> Generator["SourceContext", None, None]:
        """
        Get all the instances of
        [`SourceContext`][investigraph.model.context.SourceContext] for the
        current pipeline.

        Args:
            limit: Optionally only return this number of items (for debugging
                purposes)

        Yields:
            Generator for Source model instances
        """

        def _sources():
            yield from self.config.seed.handle(self)
            yield from self.config.extract.sources

        for ix, source in enumerate(_sources(), 1):
            yield SourceContext(config=self.config, source=source)
            if limit is not None and ix >= limit:
                return

    # RUNTIME HELPERS

    def make_entity(self, schema: str, *args, **kwargs) -> StatementEntity:
        """
        Instantiate a new Entity with its schema and optional data.

        Example:
            ```python
            def transform(ctx, record, ix):
                proxy = ctx.make_entity("Company")
                proxy.id = f"c-{ix}"
                proxy.add("name", record["name"])
            ```

        Args:
            schema: [FollowTheMoney schema](https://followthemoney.tech/explorer/)

        Returns:
            instance of StatementEntity
        """
        return make_entity(schema, *args, dataset=self.dataset, **kwargs)

    def make_slug(self, *args, **kwargs) -> str:
        """
        Generate a slug (usable for an entity ID). This guarantees a valid slug
        or raises an error. It either uses the configured dataset prefix or a
        custom prefix given as `prefix` keyword argument.

        Returns:
            A slug

        Raises:
            ValueError: When the slug is invalid (e.g. empty string or `None`)
        """
        prefix = kwargs.pop("prefix", self.prefix)
        slug = join_slug(*args, prefix=prefix, **kwargs)
        if not slug:
            raise ValueError("Empty slug")
        return slug

    def make_id(self, *args, **kwargs) -> str:
        """
        Generate an ID (usable for an entity ID). This guarantees a valid slug
        or raises an error. It either uses the configured dataset prefix or a
        custom prefix given as `prefix` keyword argument. The ID is generated
        from the arguments as a SHA1 hash (same as
        `followthemoney.util.make_entity_id`)

        Returns:
            An ID

        Raises:
            ValueError: When the id is invalid (e.g. empty string or `None`)
        """
        prefix = kwargs.pop("prefix", self.prefix)
        id_ = join_slug(make_entity_id(*args), prefix=prefix)
        if not id_:
            raise ValueError("Empty id")
        return id_

    def make_fingerprint_id(self, *args, **kwargs) -> str:
        """
        Generate an ID (usable for an entity ID). This guarantees a valid slug
        or raises an error. It either uses the configured dataset prefix or a
        custom prefix given as `prefix` keyword argument. The ID is generated
        from the fingerprint (using `rigour.fingerprints`) of the arguments as a
        SHA1 hash (same as `followthemoney.util.make_entity_id`)

        Returns:
            An ID based on the fingerprints of the input values

        Raises:
            ValueError: When the id is invalid (e.g. empty string or `None`)
        """
        prefix = kwargs.pop("prefix", self.prefix)
        id_ = join_slug(make_fingerprint_id(*args), prefix=prefix)
        if not id_:
            raise ValueError("Empty id")
        return id_


class SourceContext(DatasetContext):
    source: Source

    # STAGES

    def extract(self, limit: int | None = None) -> RecordGenerator:
        """
        Extract the records for the current source with the configured handler.
        Defaults to [`investigraph.logic.extract:handle`][investigraph.logic.extract]

        Args:
            limit: Optionally only return this number of items per source (for
                debugging purposes)

        Yields:
            Generator of dictionaries `dict[str, Any]` that are the extracted records.
        """

        def _records():
            for ix, record in enumerate(self.config.extract.handle(self), 1):
                if limit is not None and ix > limit:
                    return
                record["__source__"] = self.source.name
                yield record

        yield from logged_items(
            _records(),
            "Extract",
            item_name="Record",
            logger=self.log,
            dataset=self.dataset,
            source=self.source.uri,
        )

    def transform(self, records: RecordGenerator) -> StatementEntities:
        """
        Transform extracted records from the current source into FollowTheMoney
        entities with the configured handler.
        Defaults to [`investigraph.logic.transform:map_ftm`][investigraph.logic.transform]

        Args:
            records: Generator of record items as `dict[str, Any]`

        Yields:
            Generator of StatementEntity
        """

        def _proxies():
            for ix, record in enumerate(records, 1):
                yield from self.config.transform.handle(self, record, ix)

        yield from logged_items(
            _proxies(),
            "Transform",
            item_name="Proxy",
            logger=self.log,
            dataset=self.dataset,
            source=self.source.uri,
        )

    def task(self) -> "TaskContext":
        """
        Get a runtime task context to pass on to helper functions within
        transform stage. See [`TaskContext.emit`][investigraph.model.TaskContext.emit]

        Example:
            ```python
            def transform(ctx, record, ix):
                task_ctx = ctx.task()

                # do something, within this function use `task_ctx.emit()`
                handle_record(task_ctx, record)

                # pass on emitted entities to next stage
                yield from task_ctx
            ```

        Returns:
            The runtime task context
        """
        return TaskContext(**self.model_dump())

    def open(
        self, mode: str | None = DEFAULT_MODE, **kwargs
    ) -> ContextManager[IO[AnyStr]]:
        """
        Open the context source as a file-like handler. If `archive=True` is set
        via extract stage config, the source will be downloaded locally first.

        Example:
            ```python
            def extract(ctx, *args, **kwargs):
                with ctx.open() as h:
                    while line := h.readline():
                        yield line
            ```

        Args:
            mode: The mode to open, defaults `rb`

        Returns:
            A file-handler like context manager. The file gets closed when
                leaving the context.
        """
        uri = self.source.uri
        if not self.source.resource.is_local:
            response = fetch(uri, dataset=self.dataset)
            return response.context.open(response.content_hash)
        return smart_open(uri, mode=mode, **kwargs)


class TaskContext(SourceContext):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    proxies: dict[str, StatementEntity] = {}
    data: dict[str, Any] = {}

    def __iter__(self) -> StatementEntities:
        yield from self.proxies.values()

    def emit(self, *proxies: StatementEntity | None) -> None:
        """
        Emit Entity instances during task
        runtime. The entities will already be merged. This is useful for helper
        functions within transform logic that create multiple entities "on the
        fly"

        Example:
            ```python
            def make_person(ctx: TaskContext, record: dict[str, Any]) -> E:
                person = ctx.make_entity("Person", id=1, name="Jane Doe")
                note = ctx.make_entity("Note", id="note-1", entity=person)

                # make sure the note entity is emitted as we are only returning
                # the person entity:
                ctx.emit(note)

                return person
            ```

        """
        for proxy in proxies:
            if proxy is not None:
                if not proxy.id:
                    raise DataError("No Entity ID!")
                # do merge already
                if proxy.id in self.proxies:
                    self.proxies[proxy.id] = merge(self.proxies[proxy.id], proxy)
                else:
                    self.proxies[proxy.id] = proxy


@cache
def get_source_context(
    config_uri: Uri, source_name: str, uri: str | None = None
) -> SourceContext:
    config = get_config(config_uri)
    for source in config.extract.sources:
        if source.name == source_name:
            return SourceContext(config=config, source=source)
    if len(config.extract.sources) == 1:
        return SourceContext(config=config, source=config.extract.sources[0])
    if uri:
        return SourceContext(config=config, source=Source(name=source_name, uri=uri))
    raise ValueError(f"Source not found: `{source_name}`")


@cache
def get_dataset_context(config_uri: Uri) -> DatasetContext:
    config = get_config(config_uri)
    return DatasetContext(config=config)



"""
Extract sources to iterate objects to dict records
"""

from runpandarun import Playbook
from runpandarun.io import guess_handler_from_mimetype

from investigraph.exceptions import ImproperlyConfigured
from investigraph.model.context import SourceContext
from investigraph.types import RecordGenerator


def extract_pandas(ctx: SourceContext) -> RecordGenerator:
    play = ctx.source.pandas
    if play is None:
        raise ImproperlyConfigured("No playbook config")
    play = play.model_copy(deep=True)
    if play.read is None:
        raise ImproperlyConfigured("No playbook config")
    if play.read.handler is None:
        play.read.handler = f"read_{guess_handler_from_mimetype(ctx.source.mimetype)}"
    with ctx.open() as h:
        play.read.uri = h
        df = play.run()
        df = df.astype(object).where(df.notna(), None)
        yield from df.to_dict(orient="records")


# entrypoint
def handle(ctx: SourceContext, *args, **kwargs) -> RecordGenerator:
    """
    The default handler for the extract stage. It handles tabular sources with
    `pandas`. Custom extract handlers must follow this function signature.

    Args:
        ctx: instance of the current `SourceContext`

    Yields:
        Generator of dictionaries `dict[str, Any]` that are the extracted records.
    """
    if ctx.source.pandas is None:
        ctx.source.pandas = Playbook()
    yield from extract_pandas(ctx, *args, **kwargs)



# thanks, zavod: https://zavod.opensanctions.org/helpers/#zavod.helpers.make_address

from functools import lru_cache
from typing import Optional

from followthemoney import EntityProxy, registry
from rigour.addresses import format_address_line

from investigraph.model import SourceContext
from investigraph.util import join_text


@lru_cache(maxsize=10000)
def format_address(
    summary: Optional[str] = None,
    po_box: Optional[str] = None,
    street: Optional[str] = None,
    street2: Optional[str] = None,
    street3: Optional[str] = None,
    house: Optional[str] = None,
    house_number: Optional[str] = None,
    postal_code: Optional[str] = None,
    city: Optional[str] = None,
    county: Optional[str] = None,
    state: Optional[str] = None,
    state_district: Optional[str] = None,
    state_code: Optional[str] = None,
    country: Optional[str] = None,
    country_code: Optional[str] = None,
) -> str:
    """Given the components of a postal address, format it into a single line
    using some country-specific templating logic.

    Args:
        summary: A short description of the address.
        po_box: The PO box/mailbox number.
        street: The street or road name.
        street2: The street or road name, line 2.
        street3: The street or road name, line 3.
        house: The descriptive name of the house.
        house_number: The number of the house on the street.
        postal_code: The postal code or ZIP code.
        city: The city or town name.
        county: The county or district name.
        state: The state or province name.
        state_district: The state or province district name.
        state_code: The state or province code.
        country: The name of the country (words, not ISO code).
        country_code: A pre-normalized country code.

    Returns:
        A single-line string with the formatted address."""
    if country_code is None and country is not None:
        country_code = registry.country.clean_text(country)
    street = join_text(street, street2, street3, sep=", ")
    data = {
        "attention": summary,
        "road": street,
        "house": po_box or house,
        "house_number": house_number,
        "postcode": postal_code,
        "city": city,
        "county": county,
        "state": state,
        "state_district": state_district,
        "state_code": state_code,
        "country": country,
    }
    return format_address_line(data, country=country_code)


def make_address(
    context: SourceContext,
    full: Optional[str] = None,
    remarks: Optional[str] = None,
    summary: Optional[str] = None,
    po_box: Optional[str] = None,
    street: Optional[str] = None,
    street2: Optional[str] = None,
    street3: Optional[str] = None,
    city: Optional[str] = None,
    place: Optional[str] = None,
    postal_code: Optional[str] = None,
    state: Optional[str] = None,
    region: Optional[str] = None,
    country: Optional[str] = None,
    country_code: Optional[str] = None,
    key: Optional[str] = None,
    lang: Optional[str] = None,
) -> Optional[EntityProxy]:
    """Generate an address schema object adjacent to the main entity.

    Args:
        context: The runner context used for making and emitting entities.
        full: The full address as a single string.
        remarks: Delivery remarks for the address.
        summary: A short description of the address.
        po_box: The PO box/mailbox number.
        street: The street or road name.
        street2: The street or road name, line 2.
        street3: The street or road name, line 3.
        city: The city or town name.
        place: The name of a smaller locality (same as city).
        postal_code: The postal code or ZIP code.
        state: The state or province name.
        region: The region or district name.
        country: The country name (words, not ISO code).
        country_code: A pre-normalized country code.
        key: An optional key to be included in the ID of the address.
        lang: The language of the address details.

    Returns:
        A new entity of type `Address`."""
    city = join_text(place, city, sep=", ")
    street = join_text(street, street2, street3, sep=", ")

    # This is meant to handle cases where the country field contains a country code
    # in a subset of the given records:
    if country is not None and len(country.strip()) == 2:
        # context.log.warn(
        #     "Country name looks like a country code",
        #     country=country,
        #     country_code=country_code,
        # )
        if country_code is None:
            country_code = country
            country = None

    if country is not None:
        parsed_code = registry.country.clean(country)
        if parsed_code is not None:
            if country_code is not None and country_code != parsed_code:
                context.log.warn(
                    "Country code mismatch",
                    country=country,
                    country_code=country_code,
                )
            country_code = parsed_code

    if country_code is None:
        country_code = registry.country.clean(full)

    if not full:
        full = format_address(
            summary=summary,
            po_box=po_box,
            street=street,
            postal_code=postal_code,
            city=city,
            state=state,
            state_district=join_text(region, state, sep=", "),
            country=country,
            country_code=country_code,
        )

    if full == country:
        full = None

    address = context.make_entity("Address")
    try:
        address.id = context.make_fingerprint_id(full, prefix=f"addr-{country_code}")
    except ValueError:
        return

    address.add("full", full, lang=lang)
    address.add("remarks", remarks, lang=lang)
    address.add("summary", summary, lang=lang)
    address.add("postOfficeBox", po_box, lang=lang)
    address.add("street", street, lang=lang)
    address.add("city", city, lang=lang)
    address.add("postalCode", postal_code, lang=lang)
    address.add("region", region, lang=lang)
    address.add("state", state, quiet=True, lang=lang)
    address.add("country", country_code, lang=lang, original_value=country)
    return address


def assign_address(entity: EntityProxy, address: EntityProxy | None) -> None:
    """Assign an Address entity to a given Entity. This sets `address` property
    to the "full" property of the Address, assigns countries and sets
    `addressEntity`.

    Args:
        entity: Any Entity that can have an address assigned to it
        address: The Address entity

    Examples:
        ```python
        from investigraph.helpers.addresses import assign_address

        assign_addres(entity, address)
        ```

    Returns: None
    """
    if address is not None:
        entity.add("address", address.caption)
        entity.add("addressEntity", address)
        entity.add("country", address.countries)



import re
from functools import cache
from importlib import import_module
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from typing import Any, Callable

from anystore.types import SDict, Uri
from anystore.util import make_data_checksum
from banal import ensure_dict
from followthemoney import StatementEntity
from followthemoney.util import join_text as _join_text
from ftmq.util import (
    clean_name,
    make_dataset,
    make_fingerprint,
    make_fingerprint_id,
    make_string_id,
)
from normality import slugify

from investigraph.exceptions import DataError


def slugified_dict(data: dict[Any, Any]) -> SDict:
    return {str(slugify(k, "_")): v for k, v in ensure_dict(data).items()}


def make_entity(
    schema: str,
    id: str | None = None,
    dataset: str | None = None,
    **properties,
) -> StatementEntity:
    if properties and not id:
        raise DataError("Specify Entity ID when using properties kwargs!")

    entity = StatementEntity.from_dict(
        {"id": id, "schema": schema}, default_dataset=make_dataset(dataset)
    )
    # add the property values via this api to ensure type checking & cleaning
    for k, v in properties.items():
        entity.add(k, v)
    return entity


module_re = re.compile(r"^[\w\.]+:[\w]+")


@cache
def is_module(path: str) -> bool:
    return bool(module_re.match(path))


@cache
def get_func(path: Uri, base_path: Uri | None = None) -> Callable:
    if base_path:
        path = Path(str(base_path)) / Path(str(path))
    module, func = str(path).rsplit(":", 1)
    if is_module(str(path)):
        module = import_module(module)
    else:
        mpath = Path(module)
        spec = spec_from_file_location(mpath.stem, module)
        if spec is None or spec.loader is None:
            raise ValueError(f"Cannot load `{mpath}`")
        module = module_from_spec(spec)
        spec.loader.exec_module(module)
    return getattr(module, func)


def str_or_none(value: Any) -> str | None:
    if not value:
        return None
    value = str(value).strip()
    return value or None


def join_text(*parts: Any, sep: str = " ") -> str | None:
    return _join_text(*[clean_name(p) for p in parts], sep=sep)


def to_dict(obj: Any) -> dict[str, Any]:
    if hasattr(obj, "model_dump"):
        return obj.model_dump()
    if hasattr(obj, "to_dict"):
        return obj.to_dict()
    return ensure_dict(obj)


__all__ = [
    "make_string_id",
    "make_fingerprint",
    "make_fingerprint_id",
    "make_entity",
    "make_data_checksum",
    "str_or_none",
    "join_text",
    "clean_name",
]



"""
Inspect dataset pipeline config
"""

from anystore.logging import get_logger
from anystore.types import Uri

from investigraph.model.config import Config, get_config

log = get_logger(__name__)


def log_error(msg: str):
    log.error(f"[bold red]ERROR[/bold red] {msg}")


def inspect_config(p: Uri) -> Config:
    config = get_config(p)
    try:
        if not callable(config.extract.get_handler()):
            log_error(f"module not found or not callable: `{config.extract.handler}`")
    except ModuleNotFoundError:
        log_error(f"no custom extract module: `{config.extract.handler}`")
    try:
        if not callable(config.transform.get_handler()):
            log_error(f"module not found or not callable: `{config.transform.handler}`")
    except ModuleNotFoundError:
        log_error(f"no custom transform module: `{config.transform.handler}`")
    try:
        if not callable(config.load.get_handler()):
            log_error(f"module not found or not callable: `{config.load.handler}`")
    except ModuleNotFoundError:
        log_error(f"no custom load module: `{config.load.handler}`")
    return config



from typing import Generator

from anystore.store import get_store
from banal import ensure_dict, ensure_list

from investigraph.model.context import DatasetContext
from investigraph.model.source import Source


def handle(ctx: DatasetContext) -> Generator[Source, None, None]:
    """
    The default handler for the seed stage.

    Args:
        ctx: instance of the current `DatasetContext`

    Yields:
        Generator of `Source` objects for further processing in extract stage.
    """
    if ctx.config.seed.uri is not None:
        store = get_store(ctx.config.seed.uri)
        globs = ensure_list(ctx.config.seed.glob)
        for glob in globs:
            for key in store.iterate_keys(
                glob=glob,
                prefix=ctx.config.seed.prefix,
                exclude_prefix=ctx.config.seed.exclude_prefix,
            ):
                yield Source(
                    uri=store.to_uri(key),
                    **ensure_dict(ctx.config.seed.source_options),
                )



import os
from pathlib import Path

from anystore.model import StoreModel
from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

VERSION = "0.8.0"

DATA_ROOT = Path(
    os.environ.get("INVESTIGRAPH_DATA_ROOT", Path.cwd() / "data")
).absolute()


class Settings(BaseSettings):
    """
    `investigraph` settings management using
    [pydantic-settings](https://docs.pydantic.dev/latest/concepts/pydantic_settings/)

    Note:
        All settings can be set via environment variables (or a .env file),
        prepending `INVESTIGRAPH_` (except for those with a different alias)
    """

    model_config = SettingsConfigDict(
        env_prefix="investigraph_",
        env_nested_delimiter="_",
        nested_model_default_partial_update=True,
        env_file=".env",
        extra="ignore",
    )

    debug: bool = Field(default=False, alias="debug")
    """Enable debug mode (more error output)"""

    data_root: Path = DATA_ROOT
    """Default data directory to store archive and json exports"""

    config: str | None = None
    """Use this config.yml globally"""

    seeder: str = "investigraph.logic.seed:handle"
    """Use this seed handler globally"""

    extractor: str = "investigraph.logic.extract:handle"
    """Use this extract handler globally"""

    transformer: str = "investigraph.logic.transform:map_ftm"
    """Use this transform handler globally"""

    loader: str = "investigraph.logic.load:handle"
    """Use this load handler globally"""

    exporter: str = "investigraph.logic.export:handle"
    """Use this export handler globally"""

    cache: StoreModel = StoreModel(uri="memory:///")
    """Runtime cache"""

    store_uri: str = Field(default="memory:///", alias="ftm_statement_store")
    """Statement store for entity aggregation"""
